/*ar202-feather_2htu31d_2wifi

i.e. ARuino code serial 202 for (adafruit) feather (huzzah 8266 wtih) to htu31D (snmesors from Adafruit) and two
  sets of Wifi creds, selectable by wire link.

TO WHICH, as extension beyond ar787, is added a "how long switch closed" sensor. (ar787 already had a DIn sensor)

(previously known (c. 15-19 May 23, on pc2ndEllie) as a short-lived fork of ar787-feather...)
V e r s- see below
Started- 23 Feb 23 (ar787, the bulk of the code) (ar202, by that name, started 19 May 2023)*/

/*Forward declarations of some SRs that the compiler "couldn't see"
without these.... */

void SendWelcomeAndDocumentationToSerialMonitor();
String FloatPercentToFixedIntString(float fConvertMe, byte bShift);
void InitPinModes();
byte bSetUpSensors();
boolean boUseFirstSetofCreds();
ICACHE_RAM_ATTR void handleInterrupt();
void ConnectWiFi();
String sMassageMonitoredDIn0ForHuman(byte bTmpRaw);

/*Arduserver: A device to read sensors and make the reading available
in a page of HTML. See...
https://sheepdogguides.com/ardsrv/
for a full guide to the idea.
There is a wide range of open source code available to you there.

See https://wywtk.com//ardu/aht3inter_count.htm
  for help on interrupts, ISRs, counters, and also on using millis() for timing.
https://techtutorialsx.com/2016/12/11/esp8266-external-interrupts/
.. and...
https://randomnerdtutorials.com/interrupts-timers-esp8266-arduino-ide-nodemcu/
... also useful

*** SEE http://wywtk.com/fw/ardusftw/ar787/ar787_arduserver.htm 
 for lots of help with setting a very similar predecessor of this up ***
 Add a link between pinInterrupt and pinBlueLED to allow ar202 to take
 advantage of the clever trick built into it.
 
==========================================================================
Sections in this sourcecode... preface any of these
with "st" (Search Target) to find easily
  Design
  Changing the sensors
  I2C issues
  History
  To Dos

/*st===================================
stDesign

**Timers. Sigh.
18 May 2023, 10:11, I started trying to build in TWO "timer" mechanisms.
One was to make to possible to get rid of the long "delay()"... "Delay" is a
  blocking function. (That long delay is there to stop the program  from readiong the
  sensors/ updating the page ridiculously frequently)

The other was to provide a regular on/off/on... signal that will drive an interrupt,
  to monitor the state of the door (open?shut?) without any need for software polling.

What could possibly go wrong?

The answer:
  unlongPreviousMillis will only be reset
  at the end of each LONG delay.

  the "on/off/on/off..." of the LED (which will)
  trigger the ISP) will be achieved by using
  the modulo function on millis since unlongPreviousMillis
  was last reset.


((Design)"Timers. Sigh." sub-section ENDS)

("Design" section ENDS)*/

/*=======================================================
stChanging the Sensors

("Changing the Sensors" section ENDS*/
/*=======================================================
stI2C Issues

I2C was designed for communications for chips near one another
on a PCB.

I've had trouble even over what seemed to me "short" distances.

Pull-up resistors... too few? too many?... may be part of the
answer, particularly with "chains" of I2C devices.

For ar787 and 202 with two Arafruit HTU31Ds, I've had
SUCCESS (in short term tests) with...
  1) my "old" (at 5/23), self-made cable, as used NC.
  2) CBL 574 Adafruit 10cm
  3) CBL 519 18cm cable made by me out of two shorter cable
In all three of these three cases, the first HU31D was
   connected to the 8266 by wires soldered to the HTU31D,
   c. 5cm long.

("I2C Issues" section ENDS*/
/*=======================================================
stHistory

AFTER 21 May 2023, 15:20, before 21 May 2023 16:00- VERY significant changes
  made by moving "bits" between the then and else clauses of the big if/then/else
  in loop(). Made big difference to responsiveness to HTML requests, and at same
  time reduced sensor thrash.

18 May 2023, 10:10: things going well. Start made at adding interrupt service routine
  and flashing LED so opening of door could be monitored without software polling.
  (Will add 5 chars to m/r "report from Arduserver" string.)
AND at replacing "delay" (blocking) with a millis() driven wait. For basis of both see...
   http://wywtk.com/ardu/aht3inter_count.htm

SEE "(st)DESIGN" for some comments on a bit of complexity that could cause hassles.

---
May 23, I brought what I thought was latest/greatest code to Essex.
Wouldn't compile. Slews of error messages. (Eventually resolved. Seems I did indeed
have "latest/greatest". Problems due(?() to slight IDE v e r s i o n differences? Ended
up working in IDE 2.1.0. Updated some libraries to then-current v e r sions.))

Had to remove a semi-colon... 
   A line something like... #include "Adafruit_HTU31D.h"//for Adafruit temperature/humidity sensor.
     was #include "Adafruit_HTU31D.h";//for Adafruit temperature/humidity sensor.

I couldn't have my old filename... too long
I couldn't have a BU of the ino in the same folder.... even with a different name.
(At some point, I updated the Ardu IDE I was using. From 1.8 something (think it was) t0 2.1.0)
I couldn't have '<\p'in a string. '<\\p' substituted, and didn't, at least, throw errors.
Duh. I didn't WANT <#p>... I WANTED "</p"! Fixed 17 May 2023

("History" section ENDS*/
/*=======================================================
stTo Dos section

22 May 2023: Add message to webpage served stating current
 value in unlongFlashCycleLength, with CONCISE indication
 of what that tells you.

 Minor "to do"... add report of state of digital input to message that
 goes to serial monitor. Not important. Is already on webpage, in bits
 for computer as well as bits for humans.

Deal with... 
   //<!--ADD NOTES ABOUT THE LED WINK CODES here-->
.. at the place that comment is, and in the early code... under
"Design" maybe? Or maybe create separate user guide.txt

c. 4 Mar 23...
A PROBLEM HAS ARISEN... does this yet quite get the connection
to the wsensorsebserving routines right? It "works"... but may be
overactive. See "poor man's event handler"

We don't want to thrash the sensor reading.. but neither
do we want to leave requests for web-service unanswered for
too long. NEED TO SORT THIS OUT!!!

("ToDos" section ENDS*/

/*===================================================================
=====================================================================
Code continues from the bits at the top...

None of the following down to "end of non-deal breakers"...
is a "deal breaker". They merely put documentation I
  find useful into the compiled code.*/
String sPrgmName = "ar202-2at312wifi";                 //Filename identifying this "sketch" (program/app).
String sInstance = "ICH09w";                           //ID of which specific Arduino this was compiled for
String sHTMLTitle = "ICH09w-Arduserver by wywtk.com";  //This will be used in
  //HTML <title>, and thus affect the text on the tab in some browers.
String sVersID = "21 May 2023, 15:50-disto copy minus wifi creds";
//Version... can say more than just datestamp
//but datestamp saves you having to keep track of serial numbers.
//The following are lines which go into the HTML. The "A" set go in
//  if you are using the first set of wifi creds; the B set otherwise.
//They will be combined into a single <p>...</p> unless you include
//  some HTML inside them. A <br> would split them.
//  If you WANT them as one line (which the browser will of course
//  wrap) include a space at the end of the first line.
String sHwConfig = "23518a00";  //23517a00  was first ever sHwConfig- created 17 May 2023
//   for Adafruit Huzzah 8266 serving secure closet with 2 Wifi Cred sets, 2x HTU31's.
//   The long held dream of watching "in closet" and  "outside" closet first fulfilled
//   17th May 23, before breakfast with Brenda N. (D in Puerto Rico)
//   23517a00 also implies a simple digital input being monitored.
//23518a00 added a further DIn, monitored by an ISR. And also the blue LED made to
//   flash periodically, and that line connected to the IRS-monitored DIn.
String sFreeFormA0 =
  "  (This is part of the 'A' set of wifi creds, is hardcoded inside the Arduserver)";
String sFreeFormA1 = "(This is too... the second part.)";
String sFreeFormB0 =
  "  (This line is part of the 'B' set of wifi creds, is hardcoded inside the Arduserver)<br>";
String sFreeFormB1 = "    ... and this one is too...";
String sFreeForm0u, sFreeForm1u;
// ^ ^ ^ All of the above: "Instance Specific"
/*Things marked "instance specific" may need tweaking any time you
create a new instance of this device. (ctrl-f is a quick way to find them)

End of Not-Deal-Breakers*/

//The following doesn't "matter"... it can be true or false. It just changes
//(slightly) how much information goes to the serial monitor. Setting it false
//will not interfere with any basic functions.
const boolean boVerboseMMsgsWanted = true;
#define swVerboseMessagesOn //no ; with #define
/*You may want de-rem the above (Or rem it out, if you find it remmed at
    the moment). The code will work fine either way. Don't be afraid of
  the #define if it isn't an old friend... use ctrl-f to find the lines
  with swVerboseMessagesOn and all will be revealed!
  "sw" prefix for "switch"*/

//#define swCredBlock
/*De-rem the above to use 1st of TWO A/B sets of wifi creds. (Search on "swCredBlock"
    to learn more.) */

/*=======
This is able to work on two wifi LANs (not at the same time!)
And you need to know that it will go on your LAN at a specific IP
    address, and use a specific port to serve webpages. Both chosen by you
If you want to make a new instance of this device, you will need to put
    the wifi creds address, etc into the code. (ne set is enough, if
    you only want to use it in one place!)
YOU MUST "edit" THIS CODE to change those things. But you only need to change
    things up here at the top of the code.... E.g. the place you "set" the IP
    address has the following has "bLIPA0a=192;" in it.

The declarations for setup() and loop() are at the top of all the subroutine
    (SR) declarations. Study them, if you would understand how this works. The
    rest is "details"... which ARE available, but be clear about what's in
    setup() and loop() before you try to understand anything the details.*/

/*It is written so that you should be able to build it up, get it working, in
    stages. You don't need "everything" working before you can get ANY result.*/

/*Trying to make one of these, and can't even read the sensors, let alone get
the device to serve a web page? Your cable may be too long. Or "not the right
sort". Get the code to read the sensors working reliably before starting work
on the wifi things. (In the early "just read the sensors" state, you can see
the sensor readings in the serial monitor. (For help on "serial monitor":
  https://sheepdogguides.com/arduino/aht1serimoni.htm
I have a badly bodged 60cm cable that WORKS and a nice, neat,
commercially sources 50cm ribbon cable with crimped-on connectors which doesn't.

Start with a nice commercial SHORT I2C cable... 10cm? (Buy two, while shopping.)

The cable you want (for the Adafruit HTU31D, anyway) is sold as a "Qwiic" or
"STEMMA QT". (A "STEMMA" (no "QT") cable is SOMETHING ELSE.) Sigh. But at least 
Sparkfun and Adafruit are playing nicely together.

When you've got that working, try... shudder... cutting it in half and soldering
in a length of some other cable. I used alarm cable... the pairs weren't even 
twisted. (Twisted may be better... put the +ve and Gnd in one pair, SDA, SCL in
the other.) Cat5 cable is another candidate... and you can make TWO
"long" I2C cables from one length of Cat5, can't you?! Mayve useful...

 https://electronics.stackexchange.com/questions/438859/should-i-twist-dc-power-and-ground-wires-from-a-power-supply
 
---
ARGH!!!! udrSensor all mixed up! I have one udrSensor0, and FOUR things in it,
instead of 2x2. Sigh It all WORKS... but it needs sorting out, to make the code
ELEGANT!...
There is a "Things to do" in a section futher down the page.

---
What's with the arcane name, "r202-feather 2xhtu31d 2xwifi" ???...

This is my ARduino program ("sketch" if you prefer) number 202. It is
for (hardware) a FEATHER huzzah 8266, and
reads *TWO* (*TIMES*, two of them) HTU31D sensors, and sends a page
HTML out from a webserver (part of ar202), via Wifi.

The extra "2x" was meant to indicate that each HTU31D reads two things...
Temperature and humiduty.

Not my Best Name Ever. Sorry.

/*
*** HELP, PLEASE!!!???... If I accidentally posted a copy of this
*** code *with* my Wifi creds, please let me know where it is posted??
*/

/*
Pins used... See notes in
  SendWelcomeAndDocumentationToSerialMonitor()

---
To some extent, you can "pick and choose" what elements you
want to include in what YOU want to do.

It was designed to be a flexible starting point.

In particular, you can change what sensors the device monitors.

---
Before you worry about what bits you want, study the following,
  so that you understand what is here, what is available to you...

What it does, as is:
  It reads two HTU-31D temperature/ humidity sensors
  It passes the values to the serial monitor. (Or the display of you devising)
      (If infamiliar with "serial monitor" see
           https://sheepdogguides.com/arduino/aht1serimoni.htm
  It makes the values read available via the webserver that is also
      in the code, un the Feather Huzzah.

An effort was made to make it easy to change WHAT sensors are read.
An effort was made to make it easy to "unplug" the webserver element.

------
To adapt for you own purposes...

If you don't need I2C services, lines that can be dropped are marked. (They
  are only needed to read the sensors this was written for.
If you are reading different I2C devices, you will have to dig a bit in the code.

If you don't need the webserver element, remove or rem out all of the material
  marked as "for wifi", or "for webserver".

If you DO want the webserver element, you need to know that this is written on
  the assumption that you will use a static IP for the device. (A static IP within
  your LAN. Your router can, as usual, have a dynamic IP address.)

Still on "if you want webserver..."... you will, as is usual, have to set various
  things like what static IP address you want and wifi creds. An effort has been
  made to make it easy for you to recode the software to supply those data.
*/

//================================================================
/*====== BEGINNING OF: ---Section "Changing the sensors" ---- 

Guide to making changes to the code to deal with a change
in the sensors attached....

Be careful... do not attach the sensor until the pins involved
    have the right pinMode statements
  Study InitPinModes(); (which is called from setup()) to see
    the sort of thing needed.

To reflect the new hardware involved...
  create any constants needed for assigning pins to serving
     the sensor. Use these instead of just using the pin
     number wherever it is needed so that moving the sensor
     to a different pin is easy.
  update SendWelcomeAndDocumentationToSerialMonitor(); with the
     information about what is expected to be connected

Make relevant additions to InitPinModes(); 

---- v v v
THINK... long and hard, now, about the following. The
  more you get clear about now, the less hassle adding
  the sensor will ential...

What readings could come back from the lowest level SR
  involved in reading the sensor?

What do you want to do with them?
  a) for the needs of SendReadingsToSerialMonitor
  b) for the needs of sBuildMessage

The former is for humans to read
The latter is for computers to read, use.
 The latter should always be given something
   of a very specific format. Not least a fixed
   width format. And it MAY be as well to give
   it a string with digits only, if it represents
   a number. The FarWatchWatcher program writers
   won't thank you for returning "1e" as an error
   message relating to something that should have
   been a two digit number. Unless you are
   using hex, to return anything up to 255 as
   two characters.

^ ^ ^ ----  
Create a "read it" subroutine STUB, and a varkialbe or udr for
    communications between the rest of the code and the sensor,
    udr if the sensor does not return only a single datum.
  It should report some plausible numbers or strings,
    of the format that the final subroutine will employ.
  Create suitable "support" constants (or pre-filled udr
    elements). For instance, if the new sensor returns a tture,
    a "support constant" would hold a string indicating the
    units in which the value will be reported.
  This may involve creating a new TYPE of udr, of course.
  
Alter SendReadingsToSerialMonitor(udrSensorData0); (called 
  from loop()) to include report of what the sensor is reading.
  It will only report the "for testing" value for now, but 
  this doens't matter.
  
Alter sBuildMessage(... to add the new sensor to the reports.
  this will entail not only changes to the internals of
  sBuildMessage, but also the parameters it expects, the
  parameters passed.

NOW you can go back to the SR for reading the sensor, and
  get it to return values from the sensor, instead of the
  static "for testing" value(s) it was returning.

Job done!

==="Details" in respect of Changing the Sensors...
  Let's say that the new sensor returns a percent.
  And let's say that you have decided that it is
    allocated two bytes in the string sBuildMessage(
    will build.
  What do you do about a value of 100? We'll say that
    things in the range 89.4999 to 100.000 (inclu)
    should be reported as "99" But we will deal 
    with that OUTSIDE of the SR which reads the 
    sensor. (Example below)
  Write the SR for reading the sensor JUST to read
    the sensor, and to return SOMETHING that is an
    obvious, linked-to-the-nature-of-the-sensor
    value... e.g. in this case, it might return 100
  Deal with the limitations of how you decided the
    sensor values should be reported as follows...

  Create small SR...
    String sMassagePercentToTwoDigits(int iReadingWas){
      String sTmpL=String(iReadingsWas);
      if (iReadingsWas=100)
        {sTmp="99";}
      return sTmpL;
      }//end sMassagePercentToTwoDigits

  ... and wrap the "raw" value from the sensor
    in that when you do SendReadingsToSerialMonitor.

  Refinement of the above:
    If there is ANY chance of the sensor which
      is supposed to be returning a perecent
      returning anything other than a number
      in the range 0-100 (incl), you should
      decide on an error reporting scheme.
    As the value returned by all such "Massage..."
      SRs is a string, you could decide to return
      some letters. This would leave 00-99 free
      for returning percents.
    Or, if only numbers suit your purposes for
      some reason... that would suit the wants
      of people writing FarWatchWatcher programs
      in this case, then perhaps set aside either
      00-05 or 94-99 as error codes.
    Besides values of 100 (possible for a %
      sensor) and negative or >100 values arising
      from system malfunction, the "not a %"
      numbers could be error codes from the
      sensor... error codes you should report,
      in some form, anyway, to users of the
      Arduserver, don't you think?

The upshot of the above is that you may need
  layers of SRs!....

The sequence for reading a percent, even if
  you are able to say the sensor WILL return
  something from 0-100 (incl), might be...

     iSensorSaid=ReadSensor(pinNumber);
       //could return 0-100
     iSensor=BeHumbleOrPreProcess(iSensorSaid);
     sForHuman=MassagePercentForHuman(iSensorSaid);
     sForComputer=MassagePercentForComputer(iSensorSaid);

(sForHuman would be used, easily, in SendReadingsToSerialMonitor,
 sForComputer would be used, easily, in sBuildMessage)

What would be in the SRs?...
  BeHumbleOrPreProcess was given that name because we said
    we KNEW that iSensorSaid HAD to be in the range 0-100,
    incl. Well, Dinozzo, I have to tell you that such faith
    in hardware is touching, but I'm gong to be humble and
    do something about the fact I might be wrong...

    String BeHumbleOrPreProcess(iTmp integer){
    //This "hides" something that shouldn't have
    //  happened, but that is better than letting
    //  it cause other problems. If you want to
    //  Do Things when either problem state is
    //  encountered, instead of returning 100 and
    //  0 in the problem situations, return
    //  "special" numbers... but if you do that,
    //  you'll have to Do Things elsewhere to
    //  deal with the fact they may be returned.
    //But at least you've a simpler task as,
    //  e.g. 104, 876 and 234 have all been
    //  reduced to one "special number" which
    //  indicates that a "too high to be a percent"
    //  number was returned.
    
    if {(iTmp<0) then return 0;} else
       {if {(iTmp>100) then return 100;}} else
          {return iTmp;}}
     }// end of BeHumbleOrPreProcess

  I called this BeHumbleOrPreProcess... we've
    covered the "BeHumble". What is the "PreProcess"
    part?

    Suppose that the sensor can return the percent with
      four decimal places, e.g. 12.9876.. but you are
      happy to throw away two of them, i.e. happy to
      report 12.99 (I'm ROUNDING the value, note.)
    You might as well do that here, in your
      BeHumbleOrPreProcess SR.

  MassagePercentForHuman could turn the 12.99, to
      continue previous example, into "12.99%". Might
      as well do that here as clutter the
      SendReadingsToSerialMonitor SR with that detail.
    You don't need to hardcode the units, or how
     they are appended... that they can still come
     from the global you set up for them.

  Having a separate sForComputer might seem like overkill,
      but what about this? Suppose you wanted to put
      12.99 into the string that will be put in the
      webpage for a computer to read as 1299. It saves
      a character (I have files with 10s of thousands of
      readings from Arduservers... those unnecessary
      characters can add up!)
    Changing 12.99 to 1200 also means that it can be handled
      as an integer in any FarWatchWatcher program. 

           ====================
Relevant general advice...
  Even this "sub task" can benefit from the wisdom 
    encapsulated in the following adage:
      Design from top to bottom;
         develop from bottpm to top.

  Start by thinking carefully and in detail EXACTLY what
    result you want. What will the webpage look like
    to human readers? How will the new sensor reading(s)
    be encoded into the machine-readable line???

  Think about the road from what can be done in the
    the program to capture the raw reading from the
    sensor, what will need to happen to it along
    the way to the output.

  With those facts in hand, think (in SR terms) how
    you will get from raw to delivered.

  THEN start to code.
    Build any sub-sub-sub routines, and scratch routines
      to test they are working properly.
    Build your sub-sub routines, using, of course, the
      sub-sub-sub routines made for the purpose. Test.
    Build the sub-routines. Test. Plug them into the
      program proper!

---Section "Changing the sensors"        *** ENDS *** === */
//===============================================================



//================================================================
/*====== BEGINNING OF: ---Section "Big section: Things to Do" ---- 

ADD WINK CODES!!! M/R LINE

---
Give any instance of this an ink-on-paper label saying that a message 
is sent via serial monitor **at 9600 baud** on reset.*

---
I2C: PULL-UPS (1K) on SDA and SCL ***CAN HELP***!!!... maybe.
  They DID... once anyway.
 (Without them, readings can be bad.)

---Section "Big section: Things to Do"        *** ENDS *** === */
//===============================================================


#define pinMonitoredDIn0 12  //This pin will be set as input, and
                             //monitored for on/off
#define pinMsgsByLedFlash 0  //It would be sensible to use the
//onboard LED for this. On a Feather Huzzah, that is on D0
#define pinAlwaysLow 15   //Pin number in name as these pins are
#define pinSelectWifi 14  //  mentioned in documentation, by number
//16 didn't work for this... 15 couldn't pull it down very far. And strange
//   things happened anyway when link was in place. Wasn't ignoreed, but
//   neither did "boot with other wifi creds" happen. (And when I JUST
//   changed to using 14... all went well! (^_^)
//It is unusual to include something like the "14" in something like
//   pinSelectWifi. It degrades the efficiency of USING names.
//   In this case, I wanted to help the user
//WRITE UP: 16 not as simple as it seems. Even allowing for info at
//   Adafruit page... https://learn.adafruit.com/adafruit-sht31-d
//     -temperature-and-humidity-sensor-breakout

// no ; on a #define line
#define bBlueLEDOn 0 // the right value to use as xx in digitalWrite(pinBlueLED,xx) to turn LED off.
#define bBlueLEDOff 1

//The next two new in ar202, enhancements to ar787. They must be connected with
// a simple wire link for the "trick" they implement to work.
#define pinInterrupt 13  //This pin will be monitored by the ISR
#define pinBlueLED 2     //This pin will be made an oscillating outoput, and connected to
                         //pinInterrupt. On the Adafruit Feather Huzzah 8266 it also drives a small blue LED.

#define unlongFlashCycleLength 2400
#define unlongFlashCycleOffTime 2200
//The previous two are for use with MaybeChangeStateOfLED...
//The define the overall length of the cycle in ms, and the "off" time
//  in each cycle.

byte bBlueLEDCurrentState,bBlueLEDStateShouldBe;
//The previous two are for use with MaybeChangeStateOfLED


#include <Wire.h>             //to allow use of I2C
#include "Adafruit_HTU31D.h"  //for Adafruit temperature/humidity sensor.
//Use Tools/ Manage Libraries to install. See the following if
//  more help needed... https://github.com/adafruit/Adafruit_HTU31D

#include <ESP8266WiFi.h>  //needed to use wifi

#include "Tkb23227_1stLib.h"  //Only needed in connection with  generating
//the page of HTML that is available if you use wifi elements of this.
//Doesn't need to be "installed" in your IDE... just keep copies of...
//  Tkb23227_1stLib.h and Tkb23227_1stLib.cpp
//in the same folder as you .ino file.
//Indulge me?... This is a scrap left behind to celebrate the
//   day I made my 1st Arduino library... a step up I should have
//   made LONG ago. My EASY succes was all thanks to the very
//   VERY good introduction in...
//   https://roboticsbackend.com/arduino-create-library/
//   The following, if put in setup() demonstrated
//      that my first library, Tkb23227_1stLib.h was working...
//      Serial.print(addTwoInts(3,5));    ! ! ! (^_^)
//
//The library supplies a few funtions you can perfectly well put in
//the main page, i.e. the one you are reading!
//*** USING YOU OWN LIBRARIES is a WONDERFUL aid to Doing Things Well,
//  and brings no end of BENEFITS. Well worth the very small amount of
//  trouble to master the skill. See....
// https://roboticsbackend.com/arduino-create-library/
//... which not only has the right stuff, but is Very Well Written.

//The next is part of what you need to read the HTU31-D sensor. Mysterious?
//I think so too! But It Works. Also see the comments below teh two lines...
Adafruit_HTU31D htu = Adafruit_HTU31D();
Adafruit_HTU31D htu41 = Adafruit_HTU31D();
/*The above are actually a little like the more familiar...
String sVersID="vers 24 Feb 25-11:30";
... with which we created a variable: sVersID
  of a particular TYPE: "String"
  ... and gave it an initial value: "vers 24 Feb 23-06:00"
"htu", "htu41" are "variables" we create with these lines.
The other parts of the lines "just work", thanks to things
  in the Adafruit_HTU31D.h library.
The "initial value" fills bits of htu, htu41 with various good
  initial values. Note that it is THE SAME for both "variables",
  no distinction is made between htu, htu41*/

//Variables with "tmp" ("temporary") in the name are for short
//  term, narrow scope use. Do not expect anything in such a one to persist.
//("temperature" will be abbreviated "tture".)
String sTmp;
byte bErrCode;  //Remember: Type "byte" is UNsigned. It can hold 0-255
  //Not enough it done with bErrCode yet. More should be done, but the
  //  gaps are not stopping the program from working when all is well.

byte bMonitoredDIn0Raw;
String sMonitoredDIn0ForHuman, sMonitoredDIn0ForComputer;

int iTmp, iPortU;  //Type "int" is signed. 16 or 32 bit,
//            depending on processor
unsigned int unintPassesSinceStrt;
boolean boLEDState = true, boWeAreUsing1stSetOfCreds;
unsigned long unlongRecentMillis, unlongPreviousMillis;  //
  //unlongRecentMillis will be reasonably up to date.
  //It will be no further behind than when the main loop 
  //  was most recently re-started
int iSerialBaud = 115200;  //This is for the SerialMonitor.
//The value here, and the value chosen interactively for the
  //"baud", lower right-hand corner, Serial Monitor, must match.
  //MARK THE DEVICE with note to say "watch serial monitor, 115200"

// are obsolte, can be written out...
const byte bBlueLedOn = 0;   //These are the values for digitalWrite to the
const byte bBlueLedOff = 1;  //  output the LED is on. Which makes it on, 0 or
//1 depends on how the LED was wired to the circuit. On,off/0,1 is
//right for pinMsgsByLedFlash

boolean boBlueLedState = LOW;  //new in ar202

volatile byte bInterruptCounter = 0;  //We need exactly this data type, declared thus, pp
//Move to a part more clearly aobut config?
//https://randomnerdtutorials.com/interrupts-timers-esp8266-arduino-ide-nodemcu/
//bInterruptCounter is "private" to the needs of the ISR (it creates, in effect, a stack)
//See aslo... v  unintNumberOfInterrupts

unsigned int unintNumberOfInterrupts = 0;
//See also.. ^..  bInterruptCounter
//unintNumberOfInterrupts is the "public" tally of interrupts to have happened to date.



//Introduce a bBlueLedOnPin (consistent with other pin names)? Eschew? "pinMsgsByLedFlash"?
boolean boMakeHiToTurnOn = true;  //set true if making the pin high makes LED go on.

const unsigned long unlongLongDelay = 15444;       //Max: 4,294,967,295
//This determines, roughly, the   (commas not allowed)      222 333 444
//"idle" time between the times when the code reads the sensors, which
//I have also called the "long delay". A few things happen even during
//the long delay... the blue LED goes on/off/on./.., interrupts are counted,
//requests for HTML to the webserver parts of the code are serviced.

word un16ResultOfRemainder;//Word-type data store an unsigned number of *at least* 16 bits.
//16 bits gives from 0 to 65535... Some processors will store more than 16 bits.
word un16BigLoop=0,wBigLoopsCount=0;


byte bLedCurrentState,bLedStateShouldBe;
//The previous two are for use with MaybeChangeStateOfLED

/*=========
 "Unnecessary" frill BEGINS... but used for a reason!!!
 This will greatly simplify tweaking this code if you want to work
   with a different set of sensors on your device.
 My first Arduino "struct"!...
 Thank you https://forum.arduino.cc/t/using-struct/42793/2 */

//If sensors in system change, the inner parts of the following
//   will need changes.
//As it stands, this was badly done. But by the time I realized
//   this it would have meant re-writing too much. (I should have
//   put only ONE tture, one humidity in the structure, and used
//   TWO instances of TudrSensorData... one for each sensor. Duh.
//YOU can re-write it thus! It would be good practice for how you
//   would re-program the code to accomodate different sensors!
//Note to self: When you set about fixing this, start with what's
//   in the loop()SR
struct SudrSensorData {
  String sSensorUnitsTture;
  String sSensorUnitsHum;
  float fTture;
  float fHum;
};

typedef struct SudrSensorData TudrSensorData;

TudrSensorData udrSensorData0, udrSensorData1;  //Creates two instances...
/*The code now has two "variables" called usrSensorData0 and ..1
 Now, elsewhere in the code, I can have things like:
   TudrSensorData udrSensorData0; <create instance
   udrSensorData0.sSensorUnits1 = "testing, testing";
This was DUMB coding... see note above. The note with
"struct SudrSensorData..."

END OF this part of "Unnecessary" frill*/

//================================================================
//===START section "Big block of Wifi related items..."-----------
//If you don't aspire to having your device be a
//    webserver, you don't need these
//
// WiFi Definitions ** SOME MUST BE TWEAKED FOR YOUR LOCATION ** //
//

WiFiServer server(80);  //This says it is going to be on port 80.
/*Don't worry... we can change that later. But it needs SOMETHING.
It was late. I was tearing hair. I'd spent ages trying to get around
a "Catch-22" connected with this. After MANY hours in quest of
"switchable" creds/settings, and much progress, it seemed this issue
might have made all my work pointless. But then I found...
https://github.com/esp8266/Arduino/issues/2128#issuecomment-357142507 
SAVED MY SANITY. It told me "Set the object up using whatever
port you like... but add ([port num]) to your server.begin, and that
will give you a server on THAT port!!! HURRAH!!!
 */





//----- START of two alternative blocks defined by #ifdef swCredBlock...--
//Early in the code, there may or may not be a line like...
// (keep this remmed out always)   #define swCredBlock
//If it is NOT remmed out, we use the FIRST *pair* of creds

//CAREFUL: set BOTH bLIPA3a(and -b), AND bLIPAa(and -b)...
//                             bLIPA2a/b is sometimes 0, sometimes 1!


//WITH the swCredBlock mechanism we create two sets of two sets!!!
//Make the code in the FIRST block the one where everything is
//  neat, tidy, well documented. The second block needs less
//  documentation, etc, but should stay functionally identical.

//N.B. THIS v IS JUS THE "if"/"then"/"else" block! The
//  swCredBlock SWitch is DEFined, or not, ELSEWHERE
#ifdef swCredBlock
/*Here we supply TWO SETS of the various parameters for our WiFi
connection. Much of the code's documentation refers to what's
here as the "WiFi creds", but, as you can see, what you can set
goes WAY beyond SSID and password.
Which set you get depends on whether a connection ("link" has been made between
the pins called pinAlwaysLow and pinSelectWifi */
const char kSet1Label[] = "1st set";
char kSet2Label[] = "2nd set";
//
//Uses FIRST set if there is NO link. If logic changed, change Welcome message
//If your SSID includes spaces, you MAY need to preceed the sapces
//  with \'s. Let me know?!
const char WiFiSSIDa[] = "xxx";  //   (Change other part of creds, too,
  // ^ "Instance Specific"                  if data here changed)
  //By "instance" I mean the circumstances of a particular wifi LAN and this device's
  // connection to it. The code will need to be tweaked, and re-compiled, sent to
  // whatever Arduino you want to use for a particular place.
const char WiFiPSKa[] = "xxx";
// ^ "Instance Specific"  v v v as are the following!... v v v
//CAREFUL: set BOTH bLIPA3a(and -b), AND bLIPAa(and -b)...
//                             bLIPA2a/b is sometimes 0, sometimes 1!
const byte bLIPA0a = 192;
const byte bLIPA1a = 168;
const byte bLIPA2a = 0;
const byte bLIPA3a = 111;
const int iPortA = 222;
const byte bGate0a = 192;
const byte bGate1a = 168;
const byte bGate2a = 0;
const byte bGate3a = 1;
const byte bMask0a = 255;
const byte bMask1a = 255;
const byte bMask2a = 255;
const byte bMask3a = 0;


//*============
const char WiFiSSIDb[] = "yyy";
// ^ "Instance Specific"
const char WiFiPSKb[] = "yyy";
// ^ "Instance Specific"
const byte bLIPA0b = 192;
const byte bLIPA1b = 168;
const byte bLIPA2b = 0;
const byte bLIPA3b = 111;
const int iPortB = 222;
const byte bGate0b = 192;
const byte bGate1b = 168;
const byte bGate2b = 0;
const byte bGate3b = 1;
const byte bMask0b = 255;
const byte bMask1b = 255;
const byte bMask2b = 255;
const byte bMask3b = 0;


//------------------------
#else  //i.e. "#define swCredBlock" WAS REMMED *out* (swCredBlock NOT "defined")... and so... USE this, the second pair of cred sets
/*Here we supply TWO SETS of the various parameters for our WiFi
connection. Much of the code's documentation refers to what's
here as the "WiFi creds", but, as you can see, what you can set
goes WAY beyond SSID and password.
Which set you get depends on whether a connection ("link" has been made between
the pins called pinAlwaysLow and pinSelectWifi */
const char kSet1Label[] = "1st set";
char kSet2Label[] = "2nd set";
//
//Uses FIRST set if there is NO link. If logic changed, change Welcome message
//If your SSID includes spaces, you MAY need to preceed them with \'s. Let me know?!

//SET A.....
const char WiFiSSIDa[] = "your ssid";
const char WiFiPSKa[] = "your wifi password";

const byte bLIPA0a = 192;
const byte bLIPA1a = 168;
const byte bLIPA2a = 0;  // Care! 0... or 1?
const byte bLIPA3a = 111;

const int iPortA = 2222;

const byte bGate0a = 192;
const byte bGate1a = 168;
const byte bGate2a = 0;  // Care! 0... or 1?
const byte bGate3a = 1;

const byte bMask0a = 255;
const byte bMask1a = 255;
const byte bMask2a = 255;
const byte bMask3a = 0;

//-------------------------

//SET B....
const char WiFiSSIDb[] = "your 2nd ssid";  //_EXT
const char WiFiPSKb[] = "your 2nd wifi ssid";

const byte bLIPA0b = 192;
const byte bLIPA1b = 168;
const byte bLIPA2b = 0;  // Care! 0... or 1?
const byte bLIPA3b = 111;

const int iPortB = 2222;

const byte bGate0b = 192;
const byte bGate1b = 168;
const byte bGate2b = 0;  // Care! 0... or 1?
const byte bGate3b = 1;

const byte bMask0b = 255;
const byte bMask1b = 255;
const byte bMask2b = 255;
const byte bMask3b = 0;

#endif  //swCredBlock

/*----- The above provides for TWO alternative cred set PAIRSs
 (Four cred sets in all).
 
  1st level choice: by #ifdef swCredBlock
  2nd level choice: by state of pinSelectWifi

 (1st choice made while editing/ compiling/ uploading code,
  second choice by presence or absence of wire between pinSelectWifi and pinAlwaysLow)  END...---


 ^ End of "Instance Specific" material*/


/*HELP, PLEASE!!!???... If I accidentally posted a copy of this
code *with* my Wifi creds, please let me know where it is posted??
*/

/*The following may be "overkill", but I wanted everything that would
need tweaking to create a new instance of the device to be here near
start of the code.

The following determine...
  what the device's Local IP Address ("LIPA") will be. (It's IP in the LAN)
  what local IP addr is used for the device's gateway
  and the LAN's subnet mask. (All of this is IPv4 only)

bLIPA0,-1, and -2 determine the first three numbers in BOTH the
device's LIPA and that of the first three for the gateway.
(If you know of a LAN that allows these to be different, do tell!)

The final number for the device's LIPA is supplied with bLIPA3
The final number for the gateway's LIPA is supplied with bLIPA_Gateway
Not rocket science!

SubNet0,-1, -2 and -3 determine... can you guess?... the LAN's subnet mask.

If you want...
  Device's LIPA to be 192.168.0.123 then set...
    bLIPA0=192, bLIPA1=168, bLIPA2=0, bLIPA3=123
If you want...
  Gateway's LIPA to be 192.168.0.1 then set...
    bLIPA0=192, bLIPA1=168, bLIPA2=0, bLIPA_Gateway=1
If you want subnet mask of 255.255.255.0, then set...
  bSubNet0=255, bSubNet1=255, bSubNet2=255, bSubNet3=0*/

//Further wifi related items...
String sMsg;  //Wifi related. Remember: There are two easily muddled data
              //types: String and string. This must be String, upper case "S".

//---END of "Big block of Wifi related items"========





//=======================================================
//*******************************************************
//Here begins the subroutines of this code.

//setup() and loop() are at the bottom
//The otehrs are listed below, mostly in
//alphabetical order, using every character found
//on the first line of the block. I.e byte's "b" in
//"byte bOffer, void's "v" in "void SendWelcome..."
//
//(A few excptions are made to alphabeitcal ordering.
//     They are explained in the places they arise.)
//-------------------------------

//=======================================================
//====booleans ==========================================
//=======================================================
boolean boUseFirstSetofCreds()
//If there is NO link (which will let 14 flot HIGH,
//  use FIRST set of creds...
//Logic: The EASY route (no link) gives you FIRST set
{
  if (digitalRead(pinSelectWifi) == HIGH) /*Pulled high if
       there is NO external h/w link to 0v.*/
  { return true; }                        //no ; here
  else {
    return false;
  };
}  // end of boolean boUsingFirstSetofCreds()

//====bytes =============================================
//=======================================================
byte bOfferToServe(String sHttpMsg)  //Not *s*tring- *S*tring)
/* This routine taken "wholesale" from the excellent...  
https://arduino-esp8266.readthedocs.io/en/latest/esp8266wifi/server-examples.html
Any Good Stuff- all credit to the authors of that. (readthedocs accessed 25 Feb 23)
And then messed up by me! (T.K.Boyd, author of the code you find this in.)
See also >>> MY WEBPAGE (I need to insert link here. wywtk.com will do 'til I~ do<<<
*/
{
  // See if a client (web browser) trying to connect. Falls through, if no
  //   one queuing to get in.
  //In code that inspired this, next was: WiFiClient client = server.accept();
  //28 Feb 2023, 20:14,... a new problem arose... server had WORKED... but now suggesting Server (uc s)
  WiFiClient client = server.available();
  //WiFiClient clientB = serverB.available();
  //WiFiClient clientU;16

  //See https://www.arduino.cc/reference/en/libraries/ethernet/server.available/
  //and https://forum.arduino.cc/t/server-available-problem/117342
  //and https://forum.arduino.cc/t/wificlient-client-server-available-fails/703298/2
  //and https://reference.arduino.cc/reference/en/libraries/ethernet/server.accept/
  //The "problem"?... When, 25 Feb 23, Arduino 1.8.13, I tried "accept", I got...
  //  'class WiFiServer' has no member named 'accept'; did you mean '_accept'?
  //If I understand these things, "available" is better ANYWAY! "accept" would see
  //  clients that merely CONNECTED, and then did nothing. I am only interested in
  //  clients that send an http request, and such clients ALWAYS send stuff.
  //".available", as I understand things, only "sees" clients wanting to send something!
  //  (^_^). (Using "accept" risks(?) spending time (getting into a do... while... loop that
  //  won't end) if a client JUST connects, and then does nothing with the connection.
  //SEND MESSSAGE ABOUT THIS TO readthedocs!

  //The following won't be of any use unless a server.begin was successfully executed earlier!
  //Doing that in setup() seems permissible.

  if (client) {
    Serial.println("\n[Client connected]");
    while (client.connected()) {
      // read line by line what the client (web browser) is requesting
      if (client.available()) {
        String line = client.readStringUntil('\r');
        Serial.print(line);
        // wait for end of client's request, that is marked with an empty line
        if (line.length() == 1 && line[0] == '\n') {
          //original was... client.println(prepareHtmlPage());
          client.println(sHttpMsg);  //My version. Trivial change, as far
            //as the webserving is concerned. Just to do with different program
            //structure.(PrepareHtmlPage() was a subroutine in the original that
            //returned a type-String value)
          break;  //Provides for "early" exit from "while (client.connected())".
                  //Is there any way you could get STUCK inside that "while"???
                  // ^ ^ ^ IMPORTANT! RESOLVE!!! ^ ^ ^
        }
      }
    }

    while (client.available()) {
      // but first, let client finish its request
      // that's diplomatic compliance to protocols
      // (and otherwise some clients may complain, like curl)
      // (that is an example, prefer using a proper webserver library)
      client.read();
    }

    // close the connection:
    client.stop();
    Serial.println("[Client disconnected]");
  }
  //Make value returned meaningful!
  return 99;
}  //end of byte bOfferToServe()

//=======================================================
byte bSetUpSensors() {
  //If sensors in system change, the inner parts of the following
  //   will need changes.

  udrSensorData0.sSensorUnitsTture = "deg.C";
  udrSensorData0.sSensorUnitsHum = "% RH";
  udrSensorData1.sSensorUnitsTture = "deg.C";
  udrSensorData1.sSensorUnitsHum = "% RH";

  //The "while (true) mechanism uses the Watchdog timer to
  //   get us out of a jam. See accompanying tutorial.
  if (!htu.begin(0x40)) {
    Serial.println("Couldn't find sensor on 40!");
    while (true)
      ;  //"true"(all lower case. don't use "TRUE")
  }

  if (!htu41.begin(0x41)) {
    Serial.println("Couldn't find sensor on 41!");
    while (true)
      ;  //"true"(all lower case. don't use "TRUE")
  }

  //Make value returned meaningful!
  return 5;
}  //end of byte bSetUpSensors



ICACHE_RAM_ATTR void handleInterrupt() {
  //This is the interrupt service routine, the "ISR"
  //For help, see https://wywtk.com//ardu/aht3inter_count.htm
  bInterruptCounter++;
}  //end of handleInterrupt


//==== strings ==========================================
//----------- OUT OF sequence, but goes with next....
String sMachineReadableData() {
  String sTmpL, sBuildMachReadableData;

  //In the following, the float in fTture is converted into a string.
  float fTture = udrSensorData0.fTture;
  float fTture1 = udrSensorData1.fTture;
  float fHum = udrSensorData0.fHum;
  float fHum1 = udrSensorData1.fHum;

  //1st one is =, not += for a good reason!
  sBuildMachReadableData = FloatPercentToFixedIntString(fHum, 2);    //4.679% -> 0468  String(fHum);//
  sBuildMachReadableData += FloatToFixedIntString(fTture, 1, 4);     //12.38 -> +124, -4.444 _> -044
  sBuildMachReadableData += FloatPercentToFixedIntString(fHum1, 2);  //4.679% -> 0468  String(fHum1);//
  sBuildMachReadableData += FloatToFixedIntString(fTture1, 1, 4);
  sBuildMachReadableData += sMonitoredDIn0ForComputer;
  sBuildMachReadableData += FancyLongIntToStr(unlongRecentMillis, 5);
  sBuildMachReadableData += FancyLongIntToStr(unintNumberOfInterrupts, 5);

  return sBuildMachReadableData;
}  //end sMachineReadableData... out of sequence, as it goes with sBuildMessage

//=======================================================
String sBuildMessage(int iPasses, TudrSensorData udrSensorDataLocal0,
                     TudrSensorData udrSensorDataLocal1) {
  //^^^N.B.: "String", not "string"
  //          Both acceptable to Arduino, only String suits this job
  //Only needed for wifi features.
  //This builds the WHOLE of the message to be set off, HTTP style

  //If sensors in system change, the inner parts of the following
  //   will need changes.

  byte bErrReturned = 0;

  //Prepare the response. Various elements added by TKB to Sparkfun demo>
  //This has been done with more care than was necessary. Something
  //  much simpler, for the core text, is possible.
  //Start with the common header:
  String sToReturn = "HTTP/1.1 200 OK\r\n";
  sToReturn += "Content-Type: text/html\r\n\r\n";
  sToReturn += "<!doctype html><html lang="
               "en"
               ">\n";

  sToReturn += "<head><title>" + sHTMLTitle + "</title>\n";
  sToReturn += "<meta http-equiv=\"Content-Type\" content=\"text/html;charset=utf-8\">\n";
  sToReturn += "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0\">\n";

  sToReturn += "<body>\n<h1 style=\"color:#0000cc;\">Message from a wywtk.com Arduserver...</h1>";

  /*========== Here begins a somewhat arcane section
  //If sensors in system change, the inner parts of the following
  //   will need changes.

Besides being arcane, it is something of a dog's
  dinner 26 Feb 23.

  Before we feed the dog... some good documentation...
  
  This block adds some lines to th HTMP being built. The lines
  are comments... ordinary visitors to the page served won't
  see what's in them. But the contents are formatted to be
  read easily by other computers. See my "FarWatchWatcher" pages...
  http://sheepdogsoftware.co.uk/ssdd67.htm

  The format used is "understood" by one of my FW005 prgms:
  "FWDDS- Harvest data from HTTP server. Ver: Fork-of-31-
           Jan-19. Vers of fork:05 Mar 19."...
  ... which was watching 192.168.0.212:1202 at 26 Feb 23-
        NC Kitchen. It was returning...
  For THAT FarwatchWatcher, both the "hhhhsttthhhhstttb" line...
    (indication of format) and the next line, the line with
    data, had to be present. And the format line had to be
    the 8th line returned by the FarWatch Arduserver.

 I'm copying that (and adding to it) for ar787 so that I will
    have some FWW to use, at least until I can write a better
    one.
    
  Thing go downhill from here... but I hiope there's enough
    sense in the next bits to be useful...
    
       hhhhsttthhhhstttb  ... type data. (The "b" at the end
            was optional, for that, it seems, in older uses
            of this format. Pre-2023 uses.)
            
  Two samples of data that might be encountered...          
       0001+2450001+1971... or
       0001-0289999+3231

  hhhh: Could be any4 digit number! But I will use it here, for ar787, for %RH humidity * 100 (with 100% converted to
    99.99%)
  sttt: s: sign. ttt:digits. Here: tture in deg C * 10

  b: For ar787- ANY LETTER, either u/c or l/c. Reserved for future use in ar787 at the time
  of writing. 

  For ar787, I am adding 10 more digits, used for the latest result from millis().

  The following is a complete example of what the old FW Arduserver, the one the FWW
  could watch, might have had in the 7th, 8th and 9th lines of the webpage. (Pages
  counted as a human... or Textpad... would- 1st line: "line 1")
  N.B.... the last two lines would be visible even to casual visitors to
  the page, not hidden in comments as I generally prefer to do these days (2/23).
  
  <!--The next line must be the 8th line for some FarWatchWatchers<br>-->
  hhhhsttthhhhstttb
  0001+2450001+1271

  (Remember- at 2/23, the data is being expanded to add a further 10 digits.
  
  Enough documentation! Still poor material, but... It Should Work. (Re-write some day?)
  */

  //Insert five lines in the HTML. The first is hidden from the ordinary reader of
  //the page. The last two, the ""machine friendly" version of the data, must be visible
  //to allow old-at-2/23 FWWs to read them. WHEN THOSE FWWs have been retired, this
  //report of the sensor readings in m/r form can be dropped from the code... WHEN YOU
  //DO THAT take out the use of sTmp that exists at the moment.
  sToReturn += "<pre style=\"color:#cc0099;\">\n<!--The next line must be line 8 (1st is line'1') for some FWW s/w-->\n";
  sToReturn += "hhhhsttthhhhstttbcccccnnnnn\n";
  sTmp = sMachineReadableData();
  sToReturn += sTmp + "\n";  //sTmp used again in a moment!... don't upset contents!
  sToReturn += " latter part... dcccccnnnnn  ... ccccc changes if the device is running.\n";
  sToReturn += " d:State of DIn / ccccc:cycles / nnnnn:how long door has been open.\n";
  sToReturn += " Both ccccc and nnnnn are 'clipped' to their least significant 5 digits.\n";
  sToReturn += " Both are counts since the Arduserver was last booted.\n</pre><br>\n";
  sToReturn += "<!--See https://wywtk.com/fw/FarWOverview.htm ";
  sToReturn += "for an explanation of the previous few lines-->\n";
  sToReturn += "<!--The data line gives, twice...-->\n";
  sToReturn += "<!--  [ a) a tture (deg C x10) b) a humidity (%x100)]... in fixed width fields,-->\n";
  sToReturn += "<!--then a single character datum... not currently active, and then-->\n";
  sToReturn += "<!--a number that goes up and up and up (until it rolls over... which it will).-->\n";
  sToReturn += "<!--(The source of this last number is a Work In Progress. It WORKS... but WHAT-->\n";
  sToReturn += "<!--the number represents is has not yet stabilized! I is fixed width already.-->\n\n";
  //DON'T UPSET sTMP here. Set above. Used once more in a moment
  sToReturn += "<p>The two lines above are not meant for humans. They \n";
  sToReturn += "hold the current data in a machine-friendly form, for  \n";
  sToReturn += "<i>programs</i> which      access this device. The lines don't even \n";
  sToReturn += "need to be <i>visible</i> to the humans... except for \n";
  sToReturn += "the needs of some old software that I use to READ this page.</p>\n\n";
  sToReturn += "<p>If you have a look at the source of this page (if you are \n";
  sToReturn += "viewing it via a web browser), you will see that similar lines \n";
  sToReturn += "are also present in a hidden form. The visible-to-humans lines \n";
  sToReturn += "will be removed in due course.</p>\n\n";

  sToReturn += "<!--Second m/r data block... this one NOT visible to visitors-->\n";
  sToReturn += "<!--who access the webpage with a browser. This will remain when-->\n";
  sToReturn += "<!--I am able to retire the visible-to-humans data block.-->\n";
  sToReturn += "<!--hhhhsttthhhhstttbxxxxx-->\n";
  sToReturn += "<!--" + sTmp /*sMachineReadableData()*/ + "-->\n\n";
  //We're done with sTmp now. It becomes a general purpose, short term use, variable again.
  /*  
  sToReturn += "<br>-->";
  sToReturn += "<br>-->";
  sToReturn += "<br>-->";
  */

  /* BEGIN BIG BLOCK OF SCRAPS TO BE TIDIED UP....
    
  //The second parameter to the String function determines the number
  //  of decimal places... BUT MUST NOT BE PRESENT unless the value
  //  being converted is a floating point number.
  //  (If you do String(234,0); you get "0".)
  */
  //sTmp=sFixedLen(6,String(4.5)/*1st tture goes here*/);
  /*   if (bErrReturned!=0) sTmp="ErrFrm sFixedLen:"+String(bErrReturned);
      sToReturn += sTmp+",";//-99.99 provided for
  */
  //sTmp=sFixedLen(6,String(x.5)/*1st humid goes here*/);
  /*   if (bErrReturned!=0) sTmp="ErrFrm sFixedLen:"+String(bErrReturned);
      sToReturn += sTmp+",";//-99.99 provided for
  */

  /* 
  sTmp="xx";//sFixedLen(6,String(humidity,2));
    if (bErrReturned!=0) sTmp="ErrFrm sFixedLen:"+String(bErrReturned);
    sToReturn += sTmp+"-->\n";//100.00 provided for
  */
  //sToReturn += "";

  //    \n";
  // sToReturn += "and 3 )  in fixed width fields.-->\n";

  //"millis" is max 10 chars... I checked!

  //sToReturn += sTmp;

  //... "BIG BLOCK OF SCRAPS TO BE TIDIED UP"  *** ENDS *** ====== */

  sToReturn += "\n<br>\n<hr>\n";
  sToReturn += "\n<h2 style=\"color:#0000cc;\">What the sensors see...</h2>\n";
  sToReturn += "<p>When the sensors last returned a result, the two humidities ";
  sToReturn += "monitored were...</p>\n";
  sToReturn += "<h4>&emsp;&emsp;";
  sToReturn += udrSensorDataLocal0.fHum;
  sToReturn += " and ";
  sToReturn += udrSensorDataLocal1.fHum;
  sToReturn += udrSensorDataLocal0.sSensorUnitsHum;
  sToReturn += "\n</h4>\n";

  sToReturn += "<p>And the two temperatures monitored were...\n</p>\n";
  sToReturn += "<h4>&emsp;&emsp;";
  sToReturn += udrSensorDataLocal0.fTture;
  sToReturn += " and ";
  sToReturn += udrSensorDataLocal1.fTture;
  sToReturn += " ";
  sToReturn += udrSensorDataLocal0.sSensorUnitsTture;
  sToReturn += "</h4>\n";


  sToReturn += "<h4>" + sMonitoredDIn0ForHuman + "</h4>\n";
  /*  sToReturn += "<h4>&emsp;&emsp;";
  sToReturn += udrSensorDataLocal.fTture;
  sToReturn += " and ";
  sToReturn += udrSensorDataLocal.fTture1;
  sToReturn += " ";
  sToReturn += udrSensorDataLocal.sSensorUnitsTture;
  sToReturn += "</h4>\n";
*/

  sToReturn += "<p>Interrupts (door open) to date: ";
  sToReturn += unintNumberOfInterrupts;
  sToReturn += "</p>\n\n";
 
  sToReturn += "<p>Lowest (least sig) 5 digits of passes<br>";
  sToReturn += "through the loop since device booted: ";
  sToReturn += iPasses;
  sToReturn += "<br>\n\n<br>&emsp;&emsp;(There are roughly 24000 ";
  sToReturn += "passes per minute.)</p>\n\n";

  /*... "somewhat arcane" block ends here. Also block that will
     need changes if sensors changed.....  ENDS ================== */

  sToReturn += "\n<br>\n<hr>\n";
  sToReturn += "<h3>Visit https://wywtk.com for many reasons!</h3>\n\n";

  sToReturn += "<p>For tech support, and How You Could Make a Data Sender\n";
  sToReturn += "like the one that feeds this, visit <a href="
               "https://bit.ly/InfoByPcID"
               " \n";
  sToReturn += "target="
               "_blank"
               ">https://bit.ly/InfoByPcID</a>, (which takes you \n";
  sToReturn += "to <a href="
               "https://wywtk.com/fw/du/by-pc-id.htm"
               " target="
               "_blank"
               ">";
  sToReturn += "https://wywtk.com/fw/du/by-pc-id.htm</a>).</p>\n\n";

  sToReturn += "<p>It wouldn't be a trivial project to make one of these for\n";
  sToReturn += "yourself... and would only cost about $40, if that.</p>\n\n";
  sToReturn += "<p>You would need either admirable stubborness, or significant\n";
  sToReturn += "prior knowledge of doing things on a LAN, and making webservers\n";
  sToReturn += "visible to 'the world'... if you wanted to go  that far.</p>\n";
  sToReturn += "<p>Creating a sensor that can be read across your LAN wouldn't\n";
  sToReturn += "be a huge challenge.</p>\n\n";

  /*
  sToReturn += "<p>\n";
  sToReturn += "\n";
  sToReturn += "\n";
  sToReturn += "\n";
  sToReturn += "</p>\n\n"; */

  sToReturn += "<p>If the programmer who "
               "filled"
               " the device with code remembered/chose to\n";
  sToReturn += "update some important lines of the code first, you are looking at a page from:</p>\n\n";
  sToReturn += "<p>Device: " + sInstance + "<br>\n";
  sToReturn += "Running program: " + sPrgmName + "<br>\n";
  sToReturn += "In it's " + sVersID + " version<br>\n";
  sToReturn += "With hardware config " + sHwConfig + "</p>\n\n";
  sToReturn += "<p>------------------------------<br>\n";
  sToReturn += "If the lines below speak of where the Arduserver is or \n";
  sToReturn += "what the sensors are exposed to, bear in mind that the \n";
  sToReturn += "information may be out of date.<br>\n";
  sToReturn += "To change the nformation requires replacing the code in the Arduserver. \n";
  sToReturn += "It can be moved, and want the sensors are exposed to, can \n";
  sToReturn += "be changed without replacing the code.</p>\n\n";
  //But the message DOES change if the wifi creds, etc, selection is changed
  //  via the digital input dedicated to choosing creds is changed.
  //(But the message for set A and the one for set B could be set up to
  //  be the same message.)

  sToReturn += "<p>" + sFreeForm0u + "\n";
  sToReturn += sFreeForm1u + "\n";

  sToReturn += "\n</body>\n";
  sToReturn += "</html>\n";

  return sToReturn;
}  //end of String sBuildMessage

//=======================================================
String sMassageMonitoredDIn0ForComputer(byte bTmpRaw) {
  String sTmpL;
  sTmpL = "0";
  if (bTmpRaw == 1) { sTmpL = "1"; };  //05 Mar 2023, 15:47: This def passes to html
  return sTmpL;
}  // end sMassageMonitoredDIn0ForComputer

//=======================================================
String sMassageMonitoredDIn0ForHuman(byte bTmpRaw) {
  String sTmpL;
  sTmpL = "The digital input was LOW";
  if (bTmpRaw == 1) { sTmpL = "The digital input was HIGH"; };
  return sTmpL;
}  // end sMassageMonitoredDIn0ForHuman

//==== voids ==  ========================================
//=======================================================
void ConnectWiFi() {
  /*This, 08 Sep 18, not working well... fairly consisently
 the 8266 fails to connect to LAN when auto-booted post
 upload, even though turning off/ turning on again a
 a moment later gives... every time? often, anyway...
 a successful connection to the WiFi.
     ^ ^ ^ 
 The above was before I added 25 Feb 23, the...
    while (WiFi.status()!=WL_CONNECTED)
 ...loop. I suspect that will cure the old problem. 
 */

  /*DOES ARDUINO ALLOW "local" SUBROUTINES?
boUsingFirstSet could go here, if it does... but I had no
joy when I tried to nest a SR. Try again? Research?
*/

  /*===START section "This block might better be in a SR"---*/

  /* This looks complicated, even though it isn't, really.
 Messy? "Busy"? Maybe. It arises out of creating a way to
 have two sets of wifi creds, etc, "in" the device, selectable
 just by inserting or removing a link before booting.*/

  //Init, in hopes that "set B" will be chosen, in which case... job done! (^_^)*/
  String sWiFiSSIDLocal = WiFiSSIDb, sWiFiPSKLocal = WiFiPSKb, sSetInUse = kSet2Label;

  byte bLIPA0u = bLIPA0b;
  byte bLIPA1u = bLIPA1b;
  byte bLIPA2u = bLIPA2b;
  byte bLIPA3u = bLIPA3b;
  byte bGate0u = bGate0b;
  byte bGate1u = bGate1b;
  byte bGate2u = bGate2b;
  byte bGate3u = bGate3b;
  byte bMask0u = bMask0b;
  byte bMask1u = bMask1b;
  byte bMask2u = bMask2b;
  byte bMask3u = bMask3b;
  /*iPortU is set, used, elsewhere.*/
  sFreeForm0u = sFreeFormB0;
  sFreeForm1u = sFreeFormB1;
  /*At this point, all of the user creds, etc, things, are set to "B" Set*/

  //target for search:"23305-1900"
  //VERY STRANGE... and already investigated for some time!!! It
  //SHOULD have been possible to use boWeAreUsing1stSetofCreds here... but wasn't. Sigh
  //Compiler claimed the variable had not been declared in this scope... but it was
  //a GLOBAL variable!!! Why wasn't the assignment to it in setup() valid for this???
  //Maybe a forward declaration would solve problem?
  //Wanted to use...
  //if (boWeAreUsing1stSetofCreds) //if h/w link says "use A"...

  if (boUseFirstSetofCreds())  //if h/w link says "use A"...
                               // ...then change all to "A" set...
  {
    sWiFiSSIDLocal = WiFiSSIDa;
    sWiFiPSKLocal = WiFiPSKa;
    sSetInUse = kSet1Label;
    bLIPA0u = bLIPA0a;
    bLIPA1u = bLIPA1a;
    bLIPA2u = bLIPA2a;
    bLIPA3u = bLIPA3a;
    bGate0u = bGate0a;
    bGate1u = bGate1a;
    bGate2u = bGate2a;
    bGate3u = bGate3a;
    bMask0u = bMask0a;
    bMask1u = bMask1a;
    bMask2u = bMask2a;
    bMask3u = bMask3a;
    sFreeForm0u = sFreeFormA0;
    sFreeForm1u = sFreeFormA1;
  }  //end of if... (then)... clause. No else clause is needed.

  //---END "This block might better be in a SR"===

  //Set WiFi mode to station (as opposed to AP or AP_STA)
  WiFi.mode(WIFI_STA);

  Serial.println();
  Serial.print("Connecting to Wifi, using ");
  Serial.print(sSetInUse);
  Serial.print(" of creds. (ssid:");
  Serial.print(sWiFiSSIDLocal);
  Serial.println(")...");

#ifdef swVerboseMessagesOn
  SendVerboseMessage(bLIPA0u, bLIPA1u, bLIPA2u, bLIPA3u, iPortU);
  //To take this block out of your code, rem out the #Define swVerboseMessagesOn at the top
//  of the page
#endif swVerboseMessagesOn

  //WiFI.begin([ssid], [passkey]) initiates a WiFI connection
  //   to the stated [ssid], using the [passkey] as a WPA, WPA2,
  //   or WEP passphrase.
  //
  WiFi.begin(sWiFiSSIDLocal, sWiFiPSKLocal);
  /*Failure to connect handling should be present here... a task for later.
  pp the excellent...
  https://alselectro.wordpress.com/2016/11/29/esp8266-wifi-library-on-arduino-ide/ 
  ... the possible return values are
  WL_CONNECTED        after successful connection is established with the Access Point
  WL_IDLE_STATUS      when Wi-Fi is in process of changing between statuses
  WL_NO_SSID_AVAIL    in case configured SSID cannot be reached
  WL_CONNECT_FAILED   if password is incorrect
  WL_DISCONNECTED     if module is not configured in station mode*/

  iTmp = 0;
  //At this point, LED should already be off, but, just to be sure...
  LedTurnOff(pinMsgsByLedFlash, boMakeHiToTurnOn);

  Serial.println("Trying to find wifi of specified ssid. (Calling 'Wifi.status')...");
  while (WiFi.status() != WL_CONNECTED) {
    //Wait for preliminary connection. We are not yet connected to, or allowed
    //  to USE the LAN. So far the "doorkeeper" at the router has "said"...
    //  "You knocked? Who are you? What resources do you want? (LIPA, etc)."
    iTmp = iTmp + 1;
    //No delay HERE, as their "ought" to be... but there's one shortly
    Serial.print(".");
    //Send "Wink, wink, pause", to say: "Waiting for...
    //LedWinks(int iOnTime,int iOffTime,byte bRepeat,byte bPin, boolean boMakeHiToTurnOnL);

    LedWinks(80, 80, 2, pinMsgsByLedFlash, boMakeHiToTurnOn);
    delay(900);

    //From time to time, start a new line of ...s on the serial monitor
    if (iTmp % 18 == 0) { Serial.println(); };
  }  //Do all of that until (WiFi.status()=WL_CONNECTED)
  Serial.println();

  /*Begin block from https://forum.arduino.cc/index.php?topic=460595.0
See also https://arduino-esp8266.readthedocs.io/en/latest/esp8266wifi/readme.html
In particular, for our wants... ESP8266 as server (overview)...
   https://arduino-esp8266.readthedocs.io/en/latest/esp8266wifi/server-class.html
   And examples of same...
   https://arduino-esp8266.readthedocs.io/en/latest/esp8266wifi/server-examples.html
     Beware, though... that page sends you to the Arduino ref for THEIR "Wifi" if
     you click the "print()" item in the table of contents. Maybe that works as the 
     one in the ESPWifi??
(By the way... good news from first pg cited...
Note: if connection is established, and then lost for some reason, ESP will automatically reconnect to the last used access point once it is again back on-line. This will be done automatically by Wi-Fi library, without any user intervention.
We will use a static IP.
YES: This goes AFTER the WiFi.begin

See notes near top of code about why this has been done in such an arcane manner,
and about how it works. And the notes below.*/

  IPAddress ip(bLIPA0u, bLIPA1u, bLIPA2u, bLIPA3u);       //Instance specific
  IPAddress gateway(bGate0u, bGate1u, bGate2u, bGate3u);  //Instance specific
  IPAddress subnet(bMask0u, bMask1u, bMask2u, bMask3u);   //Instance specific

  /*EDIT- MOVE TO RIGHT PLACE!
What's above may be "overkill", but I wanted everything that would
need tweaking to create a new instance of the device to be here near
start of code.

>> CHECK... REVISE??>>
The following determine...
  what the device's Local IP Address ("LIPA") will be. (It's IP in the LAN)
  what local IP addr is used for the device's gateway
  and the LAN's subnet mask. (All of this is IPv4 only)

bLIPA0,-1, and -2 determine the first three numbers in BOTH the
device's LIPA and that of the first three for the gateway.
(If you know of a LAN that allows these to be different, do tell!)

The final number for the device's LIPA is supplied with bLIPA3
The final number for the gateway's LIPA is supplied with bLIPA_Gateway
Not rocket science!

SubNet0,-1, -2 and -3 determine... can you guess?... the LAN's subnet mask.
*/

  //Guide to parameters...
  //LedWinks(int iOnTime,int iOffTime,byte bRepeat,
  //   byte bPin, boolean boMakeHiToTurnOn)
  LedWinks(80, 60, 4, pinMsgsByLedFlash, boMakeHiToTurnOn);

  delay(900);  //Probably not necessary! But, won't hurt.
               //A pause here for the recent steps to "settle" means
               //  one fewer things to go wrong.

  iTmp = 0;
  Serial.print("Trying to config for wifi. (Calling 'WiFi.config') ");
  WiFi.config(ip, gateway, subnet);
  //end of block from forum

  Serial.print("Trying to connect to wifi. (Calling 'WiFi.status')...");
  //Now use the WiFi.status() function to check if the ESP8266
  // is connected to a WiFi network.
  //Start by requesting repeatedly until successful
  while (WiFi.status() != WL_CONNECTED) {
    // Delays allow  boolean the ESP8266 to perform critical tasks
    // defined outside of the sketch. These tasks include
    // setting up and maintaining, a WiFi connection.
    //wink, wink, pause (2 winks): "Waiting for...

    //LedWinks(iOnTime,iOffTime,bRepeat,bPin,boMakeHiToTurnOn)
    LedWinks(100, 50, 3, pinMsgsByLedFlash, true);
    delay(900);

    LedTurnOff(pinMsgsByLedFlash, boMakeHiToTurnOn);
    // Potentially infinite loops are generally dangerous.
    iTmp = iTmp + 1;
    //No delay HERE, as their "ought" to be... but there's one shortly
    Serial.print("x");
    //Send "Wink, wink, pause", to say: "Waiting for...
    //LedWinks(int iOnTime,int iOffTime,byte bRepeat,byte bPin, boolean boMakeHiToTurnOnL);

    LedWinks(80, 80, 2, pinMsgsByLedFlash, boMakeHiToTurnOn);
    delay(900);

    //From time to time, start a new line of ...s on the serial monitor
    if (iTmp % 18 == 0) { Serial.println(); };
  }
  Serial.println();
  Serial.println();
  //Final messages... (This comment's ID: cmnt23302a)
  Serial.print("Connected to WiFi! (^_^)  IP address: ");
  Serial.print(WiFi.localIP());
  Serial.print(" Serving on port ");
  Serial.println(iPortU);

}  //End void ConnectWiFi()

//=======================================================
void InitPinModes() {
  //Remember... Feather Huzzah places some contraints on the
  //  use of some GPIO pins. See...
  //https://learn.adafruit.com/adafruit-sht31-d-
  //         temperature-and-humidity-sensor-breakout
  pinMode(pinMsgsByLedFlash, OUTPUT);  //D0 For Blink test, msgs
    //pinMsgsByLedFlash, a system name, drives the red LED on the Huzzah

  pinMode(pinAlwaysLow, OUTPUT);  //Set LOW, as sink for inputs
  digitalWrite(pinAlwaysLow, LOW);

  pinMode(pinSelectWifi, INPUT_PULLUP);  //Change state to change
    //which of the two hardcoded sets of wifi creds s used.
    //  *link to 0v  -OR- *leave open BEFORE device
    //   is reset. You can't change creds without a reset.
    //See ConnectWiFi()

  pinMode(pinMonitoredDIn0, INPUT_PULLUP);  //The state of
                                            //this pin is part of what the Arduserver reports

  //Next two... new in ar202
  pinMode(pinInterrupt, INPUT_PULLUP);
  pinMode(pinBlueLED, OUTPUT);
  attachInterrupt(digitalPinToInterrupt(pinInterrupt), handleInterrupt, FALLING);
  //See https://wywtk.com//ardu/aht3inter_count.htm for help with previous

}  //end void InitPinModes()

//=======================================================
void SendReadingsToSerialMonitor(TudrSensorData udrSensorDataL0, TudrSensorData udrSensorDataL1) {
  //If sensors in system change, the inner parts of the following
  //   will need changes.

  //Don't include "Serial.print(unintPassesSinceStrt);" here...
  //  it is already part of loop():

  //A slight mystery regarding what the arguement of Serial.print can be...
  //It SEEMS there's a general rule something like...
  //  You can "mix" strings with +'s, but
  //  not a number and a string...
  //  Allowed: Serial.print(udrSensorDataL.sSensorUnitsTture+", ");
  //  Not: Serial.print(udrSensorDataL.fTture+" ");

  Serial.print(" Sensors report: Ttures ");
  Serial.print(udrSensorDataL0.fTture);
  Serial.print(" ");
  Serial.print(udrSensorDataL0.sSensorUnitsTture);
  Serial.print(", ");
  Serial.print(udrSensorDataL1.fTture);
  Serial.print(" ");
  Serial.print(udrSensorDataL1.sSensorUnitsTture);
  Serial.print(" // Humidities ");
  Serial.print(udrSensorDataL0.fHum);
  Serial.print(udrSensorDataL0.sSensorUnitsHum + ", ");
  Serial.print(udrSensorDataL1.fHum);
  Serial.println(udrSensorDataL1.sSensorUnitsHum);
}  // end void SendReadingsToSerialMonitor(

//=======================================================
void SendVerboseMessage(byte bLIPA0uL, byte bLIPA1uL, byte bLIPA2uL, byte bLIPA3uL, int iPortUL) {
  //Was it REALLY necessary to pass as parameters? Compiler was complaining "not declared
  //  in this scope" when I tried to use the globals
  Serial.print("Port: ");
  Serial.println(String(iPortUL));
  Serial.println("-----------");
  Serial.println("STARTs HERE: messages which only arise if boVerboseMMsgsWanted=true");

  Serial.print("LIPA: ");
  //In next... won't work if first byte-type variable not forced to String...
  //  to direct the call to the right overload, perhaps? (AFTER that you
  //  CAN have a mixture.
  //Following possible wheeze WON'T work!...
  //Serial.println(" "+bLIPA0u+"."+bLIPA1u+"."+bLIPA2u+"."+bLIPA3u);
  Serial.println(String(bLIPA0uL) + "." + bLIPA1uL + "." + bLIPA2uL + "." + bLIPA3uL);
  Serial.print("Port: ");
  Serial.println(String(iPortUL));
  //Serial.println(sWiFiPSKLocal);//Putting this on the Serial Monitor is
  //  probably a Bad Idea... but this is how, if you feel you need to.
  /*Serial.println("");
  Serial.println("");
  */
  Serial.println("ENDs HERE: Block of messages which only arises ");
  Serial.println("        if boVerboseMMsgsWanted=true ENDs HERE");
  Serial.println("        -----------");
  Serial.println();
}  //end SendVerboseMsg()

//=======================================================
void SendWelcomeAndDocumentationToSerialMonitor() {
  Serial.println();
  Serial.println();
  Serial.println("   (apologies for any gibberish before this. I don't know where it comes from)");
  Serial.print("Welcome to ");
  Serial.println(sPrgmName);
  Serial.println("      from https://wywtk.com");
  Serial.print("Instance: ");
  Serial.println(sInstance);
  Serial.print("Version: ");
  Serial.println(sVersID);
  Serial.print("Serial monitor needs to be run at ");
  Serial.print(iSerialBaud);
  Serial.println(" baud.");
  Serial.println();
  Serial.println("For lots of help with setting this up, see... ");
  Serial.println("http://wywtk.com/fw/ardusftw/ar787/ar787_arduserver.htm");
  Serial.println();
  Serial.println("Hardware: Feather Huzzah 8266");
  Serial.println("  2x Adafruit HTU31-D modules, part number 4832");
  Serial.println();
  Serial.println("There are notes about the Huzzah's use of");
  Serial.println("    several of the pins. For those, see...");
  Serial.println("    https://learn.adafruit.com/adafruit-");
  Serial.println("      -d-temperature-and-humidity-sensor-breakout");
  Serial.println("  Adafruit made special circumstances for 0,2,15 and 16.");
  Serial.println("  --- ar202, the code that is preducing this message,");
  Serial.println("      uses the following as indicated below.");
  Serial.println("      (I say 'D2' where Adafruit says 'GPIO 2'.)");
  Serial.println();
  Serial.println("  ---------------------------------------------------------");
  Serial.println();
  Serial.println("  N.B.: On the Feather, the pins are NOT brought out in ");
  Serial.println("    1,2,3... order in device.");
  Serial.println();
  Serial.println("  D0 is left simply driving the RED onboard LED; no other use");
  Serial.println("  D15 made an always-low input, as a 'sink' for link-options");
  Serial.println("  D14 is an input. The internal pull-up is active. Pull low");
  Serial.println("   (by connecting to GND... or to D15, which is held low) to");
  Serial.println("    use the second of the two sets of wifi creds held.");
  Serial.println("    Having no link will select the 1st set of creds, etc.");
  Serial.println("  D12 is an input. The internal pull-up is active. The state");
  Serial.println("   of this will be reported to in the webpage, both to humans,");
  Serial.println("   and as part of the m/r data.");
  //I tried using D16 as the input, but couldn't pull it low. 15 pulled 14 low,
  //no problems. (I set 14 to be an input pulled up by configuring it INPUT_PULLUP)
  //The same is true for pin 12.
  //15 is an output, set to 0v. You could use the board's 0v... but MECHANICALLY, a
  //link between 14 and 15 is easy. (They are NOT adjacent, but nearly so.)
  Serial.println("  ");
  Serial.println("  Also... to be written up more fully one day...");
  Serial.println("  from 5/23, this also uses two pins as follows...");
  Serial.println("  D2, which also drives the small blue LED on the Adafruit 8266,");
  Serial.println("     is configured as an output. It goes high/ low/ high/ low");
  Serial.println("     repeatedly.");
  Serial.println("  D2 should be connected to D13, through a switch,");
  Serial.println("     which is an input... an input watched by and interrupt service");
  Serial.println("     routine. The switch might, for instance, be connected to a door,");
  Serial.println("     with the switch OPEN when the door is CLOSED.");
  Serial.println("  The ISR counts edges, and so the count will tell you how long");
  Serial.println("     the door has been open!");
  Serial.println("  ");
  Serial.println("  ");
  Serial.println("       Think, though: If you use the RED LED to send msgs about");
  Serial.println("       the state of the device, that signal could be turned");
  Serial.println("       into an AUDIBLE message, too. (Multiple 'messages'");
  Serial.println("       can be achieved with different PATTERNs of flashing.)");
  Serial.println("  ");
  Serial.println("  ");
  Serial.println("  TX/RX left for just their roles connecting to the IDE,");
  Serial.println("     including communication with the Serial Monitor.");
  Serial.println("  D4, D5: Only used for I2C SDA/SCL... Feather's default");
  Serial.println("  pins for these signals.");
  Serial.println();
  Serial.println("Software you will need if you wish to recompile the code...");
  Serial.println("...(In addition to basic libraries for the board):");
  Serial.println("  Adafruit_HTU31D.h: For the Adafruit temperature/humidity");
  Serial.println("     sensor. Use Tools/ Manage Libraries to install. See ");
  Serial.println("     https://github.com/adafruit/Adafruit_HTU31D");
  Serial.println("     if more help needed.");
  Serial.println("  ESP8266WiFi.h: For the wifi");
  Serial.println();
  Serial.println("Configuration:");
  Serial.println("  A number of things in the code need changing for your");
  Serial.println("     requirements. All such are flagged 'Instance Specific'");
  Serial.println("     E.g. the two SETS of wifi creds 'in' the device.");
  Serial.println("  (But there are also link-selectable options in some cases...");
  Serial.println("     E.g. which set of the two wifi creds 'inside' the device");
  Serial.println("        are used.");
  Serial.println("  Connecting pin 14 to pin 15 will cause the device to use");
  Serial.println("     the second set of wifi creds.");
  Serial.println();
  Serial.println("-----------");
  Serial.println("The cable between the sensors is CRITICAL. Keep short.");
  Serial.println("Twisted pairs. If device isn't working, it is probably the");
  Serial.println("cable... especially if the near sensor can be seen, but");
  Serial.println("not the distal one.");
  Serial.println();
  Serial.println("-----------");
  Serial.println("Wink codes... (finish descriptions in due course...)");
  Serial.println("Incomplete list... in the order they may arise...");
  Serial.println("  3 flashes, then pause: working on an early part of connecting to wifi");
  Serial.println("  2 flashes, then pause: working on a later part of connecting to wifi");
  Serial.println("  LED on most of the time, with brief offs: Connected. Going around main loop.");
  //<!--ADD NOTES ABOUT THE LED WINK CODES here-->
  Serial.println();

}  //end void SendWelcomeAndDocumentationToSerialMonitor(

void MaybeChangeStateOfLED(){
//Yes... you could, probably should, write this to use
//global variables less.
//See notes in body of code. They cover what requirements
//  must be met, if you want to use this subroutine.  

if (un16ResultOfRemainder<unlongFlashCycleOffTime)
 {bLedStateShouldBe=bBlueLedOff;}
else
 {bLedStateShouldBe=bBlueLedOn;};

if (bLedCurrentState!=bLedStateShouldBe)
 {digitalWrite(pinBlueLED,bLedStateShouldBe);
  bLedCurrentState=bLedStateShouldBe;};
};//end MaybeChangeStateOfLED()


//=======================================================
//== ++++ ==  setup() AND loop().... section  === ++++====
//
//  ORIGINALLY... bef 21 May 2023...
//  The declarations of setup() and loop() were at the top of
//    all the SRs. The rest were in a following block, where
//    they are in alphabetical order, except as explained there

//=================================  setup() ==========

void setup() {

  unintPassesSinceStrt = 0;

  InitPinModes();  //SEE NOTES in SendWelcomeAndDocumentationToSerialMonitor

  Serial.begin(iSerialBaud);  //Get ready to send things to serial monitor.
  //"Serial monitor"?. See https://sheepdogguides.com/arduino/aht1serimoni.htm
  while (!Serial) { delay(10); }  //Wait for serial port to finish opening
  delay(200);// Give system a further chance to "settle". (Messy kludge)
  Serial.println();
  Serial.println();
  Serial.println("Hello world... now wait a moment. (Less than 5 seconds)");
  delay(2000);// Give reader a chance to see the output.
  //Don't try to write to the serial monitor too soon.
  //What you try to send may get lost!
  //
  //When working on the code, sometimes pressing the "reset" button on
  //the Arduino AFTER uploading the newest code will "fix" problems with
  //output that should be appearing in the serial monitor.

  SendWelcomeAndDocumentationToSerialMonitor();  //Details are in SR

  bErrCode = bSetUpSensors();

  //You may still get some gibberish in the serial
  //   monitor before the message you should get appears.
  //   This is something I've stuggled with, unsuccessfully,
  //   (obviously) for YEARS! Sigh. Anyone know the answer??
  //
  //NEEDS MORE WORK, especially in respect of what is returned
  //  in bErrCode, and what is then done with it!

  bLedCurrentState=bBlueLedOff;//This is the "LED" that does a steady
  //wink to give a pulse stream that, when the door is open, is
  //fed to the input watched by the ISR so that we can tell how long
  //the door is open, when. There doesn't have to be an LED on the
  //line, but it is worth having, just see the monitor is working.

  Serial.print("SetupSensors reported (currently meaningless.. this is a placeholder): ");
  Serial.println(bErrCode);

  //Do I want "did it set up okay?" reports here?? Can I tell??

  boWeAreUsing1stSetOfCreds = (boUseFirstSetofCreds());
  //(I wanted to use the result from the function again later, too, but
  //for some reason FAILED. See the rem that has "target for search:"23305-1900"
  //in it.

  if (boWeAreUsing1stSetOfCreds) { iPortU = iPortA; }  //no ; here
  else {
    iPortU = iPortB;
  };

  server.begin(iPortU);  //Part of creating the webserver.
  //Essential "ingredient" of webserver code. Remembering
  //that was last "gotcha" of big 2/23 effort. When, yet AGAIN,
  //I attempted to access the server this code should create,
  //I did so with NO feeling of anticipation! I'd been worn
  //down to the point of expecting "Connection timed
  //out... cannot  connect"!
  //
  //But! Never undaunted, Mr Murphy stepped up and immediatley
  //caused a weird seize-up of the PC I was using. It did't take
  //a full re-boot to get running again, but it was a scary moment!
  //(Task Manager sees the Arduino IDE as "Java(TM) Platform...", by
  //the way.)

  //
  //NEEDS WORK: *** Can we test that the "begin" was successful??

  ConnectWiFi();  //only needed if you need the wifi service

  //New in 202... set things up for timing first "long delay"
  unlongRecentMillis = millis();
  unlongPreviousMillis = unlongRecentMillis;

  bBlueLEDCurrentState=bBlueLEDOff;

}  //end setup()  -----------------------------------------------



//==================================  loop() ==========
void loop() {
  unlongRecentMillis = millis();  //This should not be updated elsewhere
  unintPassesSinceStrt = unintPassesSinceStrt + 1;

  //Next two take care of flashing blue LED. See ar203 for details.
  //Toggle state of pinBlueLED, if the time has come.
    //if the frequency of the winking isn't MUCH higher than
    //the frequency of completions of the "long delay", the
    //regularity of the pinBlueLED winking will suffer. But even
    //then it's not the end of the world, unless you want to know
    //how long the door was open very accurately. 
  un16ResultOfRemainder=unlongRecentMillis % unlongFlashCycleLength;
  MaybeChangeStateOfLED();

  //Perhaps re-introduce?...
  //  if ((unintPassesSinceStrt % 1000)==1){Serial.print("+");};//bf
  //  if ((unintPassesSinceStrt % 20000)==1){Serial.println();};//bf


  //======================================
  //The following "if" has a then and an else clause. The else is huge.
  //The "else" clause is executed when the "long delay" has passed. It is
  //  visited infrequently, but it is where the "meat" of the program is.

  //The "then" clause is given as few tasks as possible, but of course some
  //  things need to happen frequently. Hence the if/then/else structure!
  
  if ((unlongRecentMillis - unlongPreviousMillis) < unlongLongDelay) {
     //Here begins the "then" block. It happens often while we wait for
     //  the "long delay" to pass. All of the sensors, except the "door open"
     //  sensor, are only read once per "long delay". The "door open"
     //  sensor is on an ISR, and though new pulses from the oscillator (winking
     //  LED on pinBlueLED) are tallied, the count isn't used until during
     //  the else clause.

     //Deal with "message" from ISR....
      if(bInterruptCounter>0){
 
      bInterruptCounter--;
      unintNumberOfInterrupts++;
 
      Serial.println();
      Serial.print("An interrupt has occurred. Total: ");
      Serial.println(unintNumberOfInterrupts);     
     }//END OF "deal with "message" from ISR...."

     sMsg = sBuildMessage(unintPassesSinceStrt, udrSensorData0, udrSensorData1);

    //From here until place marked: WiFi related...
    //------ HERE BEGINS....
    //A block where partly we are wasting time... but also,
    //while we do that, we are checking whether a client
    //has requested service.
    //Call it a poor man's event handler...

    //IS THIS NEEDED??? In particular the repeated Wifi.Connect()s
    // entailed.

    //---- sub-unit...
    //Check if a client has connected, and if one has, deal with
    //  sending the webpage to it...
    bErrCode = bOfferToServe(sMsg);  //Related to using wifi to serve pages
    // ^ error code returned is meaningless as yet.
    //delay(2500);<This was how ar787 implemented the "long delay"
    //ar202 uses unlongPreviousMillis / unlongRecentMillis to
    //   implement it

    /*//---- sub-unit...
  //Check if a client has connected, and if one has, deal with
  //  sending the webpage to it...  
  bErrCode=bOfferToServe(sMsg);//Related to using wifi to serve pages
  // ^ error code returned is meaningless as yet.
  LedTurnOn(pinMsgsByLedFlash,boMakeHiToTurnOn);
  delay(90);
  LedTurnOff(pinMsgsByLedFlash,boMakeHiToTurnOn);
  //Can stay off now, until something else wants it
  //  on again (or we come back to line above it again.)
  delay(500);

  //---- sub-unit...
  //Check if a client has connected, and if one has, deal with
  //  sending the webpage to it...  
  bErrCode=bOfferToServe(sMsg);//Related to using wifi to serve pages
  // ^ error code returned is meaningless as yet.
  LedTurnOn(pinMsgsByLedFlash,boMakeHiToTurnOn);
  delay(90);
  LedTurnOff(pinMsgsByLedFlash,boMakeHiToTurnOn);
  //Can stay off now, until something else wants it
  //  on again (or we come back to line above it again.)
  delay(500);

  //---- sub-unit...
  //Check if a client has connected, and if one has, deal with
  //  sending the webpage to it...  
  bErrCode=bOfferToServe(sMsg);//Related to using wifi to serve pages
  // ^ error code returned is meaningless as yet.
  LedTurnOn(pinMsgsByLedFlash,boMakeHiToTurnOn);
  delay(90);
  LedTurnOff(pinMsgsByLedFlash,boMakeHiToTurnOn);
  //Can stay off now, until something else wants it
  //  on again (or we come back to line above it again.)
  delay(500);

  //---- sub-unit...
  //Check if a client has connected, and if one has, deal with
  //  sending the webpage to it...  
  bErrCode=bOfferToServe(sMsg);//Related to using wifi to serve pages
  // ^ error code returned is meaningless as yet.
  delay(500);

  //... END OF block where we waste time AND watch for clients
  */
    
  }  //This is end of the long "then" of "If we have not reached end of long delay"
  else {
    /*This block executes when...
        unlongRecentMillis - unlongPreviousMillis...
           i.e. the time since the last time we were here... is...
               >= unlongLongDelay
    */

    //Here begins section for reading the sensors....

    bMonitoredDIn0Raw = digitalRead(pinMonitoredDIn0);

    sMonitoredDIn0ForHuman = sMassageMonitoredDIn0ForHuman(bMonitoredDIn0Raw);
    sMonitoredDIn0ForComputer = sMassageMonitoredDIn0ForComputer(bMonitoredDIn0Raw);

    //If sensors in system change, the following will need
    //  to be reconsidered. These are HTU31 specific.
    sensors_event_t humidity, temp;  //Specific to using HTU31 sensors
    sensors_event_t humidity41, temp41;

    //Following also HTU31 sensor specific
    htu.getEvent(&humidity, &temp);        //You can't(?) insert space in this line
    htu41.getEvent(&humidity41, &temp41);  //   to make it "line up" with this.
    //I had an old note from somewhere that said spaces in the lines
    //above would(!) stop them working. Very odd. But that
    //was what the note said!
    //
    //If sensors in system change, the internals of the following will
    //     need changing.
    //
    //----
    //Would it make sense to create a "FillUdrSensorDataRecord" SR?
    //"UdrSensorDataRecord" A User Defined Record holding Sensor Data
    //.... not just recent readings, but also things like the units used.
    //
    //Some parts of .fTture=temp that don't change are filled elsewhere
    udrSensorData0.fTture = temp.temperature;
    udrSensorData1.fTture = temp41.temperature;
    udrSensorData0.fHum = humidity.relative_humidity;
    udrSensorData1.fHum = humidity41.relative_humidity;

    //- - - - END "Reqd Sensors" section

    //Provide proof of life and local report of readings...
    Serial.print(unintPassesSinceStrt);
    SendReadingsToSerialMonitor(udrSensorData0, udrSensorData1);

    
    //RETHINK?....
    //Report a complete pass through the loop with a flash of the LED.
    //When a serial monitor is attached, the line of data arising
    //  during a pass will suffice for that channel.

    LedTurnOff(pinMsgsByLedFlash, boMakeHiToTurnOn);
    delay(120);
    LedTurnOn(pinMsgsByLedFlash, boMakeHiToTurnOn);

    //And finally within the else block that happens when we HAVE reached the end
    //  of the long delay...
    unlongPreviousMillis=unlongRecentMillis;
  }  //This is end of the long "else" of "If we have not reached end of long delay"


}  //end loop() -----------------------------------------
//======================================================= End of the whole page.